
/*****************************************************
 *    Speedy Make
 *    (c) 2006 by Kim Haskell &  Denis Sureau
 *	
 *    http://www.speedymake.com
 *
 *    Freeware under the GNU GPL 2.0 Licence.
 *    Requirements: 
 *   - The Scriptol compiler to rebuild the executable.
 ******************************************************/
 

include "libphp.sol"
include "path.sol"
include "tools.sol"
include "smklib.sol"
include "libsax.sol"
include "libdom.sol"

include <scriptol/sax.hpp>

text makeName
dom make
array tagList
array sources
array objects
array actions

dict excludeList			// pair: exclude tag name and tag of list of files 		
dict actionList
dict sourceLists
array commName
array commData
text sourcesString
text objectsString
text libsString
text upperAction = ""
int counter

xml deptree 
/xml

boolean NORUN 		// Test only.
boolean STOPFLAG	// Exit on first error encountered.

enum INTERNAL_COMMAND, EXTERNAL_PROGRAM

void howto()
	print "Speedy Make 1.3 (c) 2006 Kim Haskell - http://www.speedymake.com"
	print "Command:"
	print "  smake [options]"
	print "  smake [options] makefile.sm"
	print "Options:"
	print "  no option   use the default file makefile.sm"
	print "  -h          display this help."
	print "  -b          rebuild all sources."
	print "  -t          test the makefile, do not execute commands."
	print "  -i          interrupt, stop at first error."
	print "  -v          verbose, display more info."
	print "  -s          silent, do not print commands."
	print "  -name=\"value\"  assign a value to the variable \"name\"."
	print "  -name       execute the action \"name\"."   
	exit(0)
return



/**
 *	Verify Files
 *	For each source file, check if it exists.
 */	 

int verifyFiles(text dirname)
	int count = 0
   
	for text t in sources
		if t.find(dirname) = nil let t = Path.merge(dirname, t)
		if not file_exists(t) 
			echo "\"",  t , "\"", " not found.\n"
			count + 1
		/if 
	/for	 
return count




/**
 *	Get Content
 *	Returns content of a tag (usually the list of sources) as a text.
 */	 

text getContent(text name)

	text content

	make.reset()	// points out the "makefile" tag
	make.down()		// enter it
    
	while make.found()
		if make.getElementName() = name
			content = make.getData()
		return content
	/if
	let make.inc()      

return ""



/**
 *	Create List  
 *	For a tag that holds a list of files,
 *	gets the list of sources, puts it into an array.
  */	 

array createList(xelement xe)
	text data = xe.getData()    // get it from the dom
	array srcs = explode(data, " \n\t")
	text srcString = data.trim()
	scan srcs  let srcs[] = srcs[].trim()
	sourceLists[xe.getName()] = srcs
return srcs


/**
 *	Get Tag Content
 *	Retrieve the content of a tag
 *	previously expanded with included files
 */
 
array getTagContent(text name)
	array clist

	make.reset()
	make.at(name)

	text content = make.getData()
    if content <> nil 
		clist = content.split(" \t\n")
    /if	
 
return clist	    

/**
 *	Add Program
 *	Add a command to the list of programs to execute
 */	  

void addProgram(text prog)
	commName.push("run")
	commData.push(prog)
	//print "ADD PROG RUN", prog
return	


/**
 *	Subst Action
 *	replaces a variable by its content
 *	- by a list is the symbol is $
 *	- create a command for each element if the symbol is *
 */	    

		
void expandVariable(text program)
	array tokens
	text srcname
	text extension
	text objString
	int pos
	int start
	int size = sources.size()
	int len = program.length()
	boolean flag = false	// set when variable has content or no variable used

	while pos <> nil
	
		pos = program.find("$")
 
		if pos = nil 
			pos = program.find("*", pos)
			if pos = nil
				addProgram(program) 
				break
			/if	
			start = pos
			srcname, pos, extension = extractName(program, pos + 1)
			sources	= sourceLists[srcname]
			if sources = nil return
					
			if extension <> nil
				if sources.size() = 0	break	// nothing to do
				objects, objString = changeExtensions(sources, extension)
			else
				sources = removeCompiled(sources)
				if sources.size() = 0	break	// nothing to do	
			/if
			for int i in 0 -- size
				text comm = program
				text name = objects[i]	
				comm [start .. pos] = name
				addProgram(program)
			/for	
			return	// only one subst of this kind allowed
		/if
		
		// A $ prefixed variable was found
			
		start = pos		
		srcname, pos, extension = extractName(program, pos + 1)

		sources = getTagContent(srcname)	// return the content of the tag
		if sources = nil break
		
		//print "UNCOMP"	
		//sources.display()
		
		if extension = nil		// original sources
			sources = removeCompiled(sources)
			if sources.size() = 0	break	// nothing to do
		else
			if sources.size() = 0	break	// nothing to do
		/if
		objects, objString = changeExtensions(sources, extension)	
		
		// Nothing to process, skip, don't add the command 
		if objString <> nil
			flag = true
		/if	
		program [start .. pos] = objString
		 
		pos = program.find("$")
		if pos = nil
			if flag = true 
				addProgram(program)
			/if	
			break
		/if	
			
	/while		
	

return 


/**
 *	Process One Tag (action or variable)
 *	Builds a command from a tag
 *	Argument:
 *	- the tag to process
 */	

void parseOneTag(text name)
	array srcs
	text content
	xelement xe 
	int intag // number of file in the original tag
   
   	make.reset()	// points out the "makefile" tag
   	xe = make.at(name).getElement()   	
   
	text prog = xe.nextName()   
   
   	if prog	<> "action" return

	text comm = xe.getValue("action")
	if DEBUG print "Parse", name, "for", comm
	if comm in { "parse", "build", "none", "" , "exclude"}
   		srcs = createList(xe)	// make an array of filesname from the content
   		intag = srcs.size()
   		echo intag, " file" , plural(srcs.size()), " in \'", xe.getName(), "\'"
   		//srcs.display()
   		if comm
		= "parse": 
			srcs  = srcs | parseList(srcs)	// add included files
			print ",", srcs.size() - intag ,"to include."
		= "exclude":
			text target = xe.getValue("target")
			if target = nil
				print "."
				print "Error, target required in", prog
			else	
				excludeList[target] = prog
				print ",", srcs.size() ,"to exclude."
			/if	
		else	
			print "."
		/if
		
  
		
   		//srcs.display()
   		content = srcs.join(" ")
   		xe.setData(content)
   		if DEBUG print "All sources:", xe.getData()
	   	
	/if   
   
	counter + 1
   
return



/**
 *	Parse Tags
 *	Parse the whole XML document.
 *	Process each tag according to the "action" attribute.
 */	  

void parseTags()
	counter = 0
	
	if DEBUG print "Parse Tags"
	
	for text name in tagList
		parseOneTag(name)
	/for
	if counter = 0  print "Nothing to do..."
	
return


/**
 *	Exclude
 *	Parse the whole XML document.
 *	Process exclusion according to the list of sources and target
 */	  

void exclude()

	array sourceData
	array targetData
	xelement xes
	xelement xet
	
	if DEBUG print "Exclude"

	if excludeList.empty()
		feedback("No file to exclude...")
		return
	/if	
	
	for text source, text target in excludeList
   		xes = make.at(source).getElement()
   		if xes = nil continue
		xet = make.at(target).getElement()
		if xet = nil continue      	
		
		sourceData = contentToArray(xes.getData(), nil)
		if sourceData.size() = 0 continue
		targetData = contentToArray(xet.getData(), nil)
		if targetData.size() = 0 continue
		
		for text x in sourceData
			int pos = targetData.find(x)
			if pos <> nil let targetData[pos .. pos] = nil
		/for
   		text content = targetData.join(" ")
   		xet.setData(content)
   		
	/for

	
return


/**
 *	Make Internal Command
 *	Convert an XML action to command for the emake intepreter,
 *	push name of the command and body on the stack.
 */	

void makeInternalCommand(xelement xe)
	
	text comm = xe.getValue("action")
	text content = xe.getData()

	if comm 
	=	"display" :
		commName.push("display")
		commData.push(content)
	/if
	
return

/**
 *	Make External Program
 *	Convert an XML action to a system command
 *	add content to program name or
  */	

void makeExternalProgram(xelement xe)

	text content = xe.getData()
	text name = xe.getName()
	expandVariable(purge(content))

return



/**
 *	Build Command List
 *	Process dependencies of actions in the makefile, 
 *	recursively from the top to deeper ones.
 *	- if the tag is a terminal, build a command
 *	- else call the function with each item in the list  
 */   

void buildCommand(text actionName)
	array innerList
	text content
	text ename
	xelement currelem
	
	// action already in list
	if actionList[actionName] <> nil return

	make.reset()
	make.at(actionName)
	currelem = make.getElement()
	
	if currelem = null
		print "Err", actionName, "tag not found."
		return 
	/if

	if DEBUG
		print currelem.getName(), "tag, attempting to make command:"
	/if	
	
	// check if the tag is a terminal
	
	if DEBUG print "Terminal tag, action=", currelem.getValue("action")
	
	if currelem.hasAttribute("action")
		text val = currelem.getValue("action") 
		if val.compare("run") = 0
			makeExternalProgram(currelem)
			return
		/if
		makeInternalCommand(currelem)	
		return
	/if	
	
	// not a terminal, process the content

	if DEBUG print "Non terminal tag..."
	content = make.getData()
	innerList = contentToArray(content, "")
	if innerList.empty() return
	if DEBUG print innerList
		
	for ename in innerList
		if actionList[ename] <> nil continue
		buildCommand(ename)
	/for		
	
	if DEBUG
		print "List of commands:"
		commData.display()	
	/if
	
	if  "run" not in commName
		print "Nothing to do."
		exit(1)
	/if
   
return	
	
	
/**
 *	Process Binary
 *	Get the date of the file to build if exists
 *	else the date is zero.
 *	Assign the date to included.buildDate
 */    	
 
void processBinary(xelement xe)

	text prog = xe.getData()
	prog = prog.trim()
	setDate(0)

	if BUILDALL = false
		if file_exists(prog)
			setDate( filetime(prog) )
		/if
	/if

	if VERBOSE
		if getDate() = 0
			echo " - first build"
		else	
			text d = date("c", getDate())
			echo " - last build "  + prettyDate(d)
		/if	
	/if			
	
return 
	
/**
 *	Get upper action, get dependencies,
 *	including variables in content (not in attributes)
 *	Take the content of each tag providing it is not terminal
 *	build a list of the content, add this list to a main list
 *	if the name of the tag is in this list, it is not upper
 *	returns:
 *		- candidates the list of all tags  
 */  

void getUpper()
	text ename
	text content
	array candidates = {}
	array localList = {}
	xelement currelem
	boolean bTag = false
	
	make.reset()
	text name = make.getValue("name")
	echo "Processing \"", name, "\" (", makeName, ")\n" 
    
	// parse the top XML level
	make.down()

	if not make.found()
		print "Empty makefile..."
		exit(0)
	/if

	while make.found()
		ename = make.getElementName()	// get the element

		if ename = nil continue		
		tagList.push(ename)

		// processing next tag
		currelem = make.getElement()
		if VERBOSE
			echo "* ", ename
		/if	

		// if internal command, skip it
		if currelem.hasAttribute("action")
			text cact = currelem.getValue("action") 
			if VERBOSE
				echo " (", cact, ")"
			/if	
			if cact = "build"
				processBinary(currelem)
				bTag = true 
			/if
			if VERBOSE print
			continue
		/if	
		
		if VERBOSE print

		// now processing a non-terminal tag

		candidates.push(ename)	// add element to list as already parsed
		content = make.getData()	// get content
		localList = contentToArray(content, "")	// build a list

		if localList.size() > 0
			if VERBOSE
				print localList.size(), "tags found."
				if DEBUG		
					print "List of tags:"
					scan localList print localList[]
				/if	
			/if			
		else
			print "No action found."
		/if
				
		// scan this local list, if tag name already in main list
		// remove it 
		scan localList
			text item = localList[].toText()
			int index = candidates.find(item)
			if index <> nil
				candidates[index] = nil
			/if
		/scan
		
	let make.inc()

  	make.up()

	if candidates.empty()
		print "No top action found, please create one..."
		exit(0)
	/if
	if candidates.size() > 1
		print "Several top actions found, please make only one..."
		scan candidates let print candidates[]
		exit(0)
	/if

	upperAction = candidates.pop()
	if VERBOSE
		print "Top action:", upperAction
	/if	

	if not bTag
		print "Error, no build action, nothing to build."
	/if

return


/**
 *	Test Conformity
 *	Check for the presence of one-letter tags that are forbidden.
 */	 

void testConformity()
	
	if VERBOSE	echo "Checking for conformity..."
	make.reset()
	make.down()
	while make.found()
		text name = make.getElementName()
		if name.length() = 0 continue
		if name.length() < 2
			print
			echo "\"", name, "\""
			print " only one letter is not a valid element name..."
			exit(0)
		/if
	let make.inc()
	feedback(" ok.")      
return



/**	
 *	Execute Commands
 *	Execute the list of commands
 *	in the array 
 */	 

void executeCommands()
	int res = 0
	text name
	text content
	text type
	
	if NORUN return
	
	scan commName, commData
		if commName[]
		= "display"
			text t = commData[]
			for text x in t
				if x = "\t" continue
				echo x
			/for	
			print
		
		= "run"
			if  NORUN return
			//print "RUN ", commName[], commData[]		
			res + exec(commData[])
			if STOPFLAG
				if res <> 0 let die("Stopped on error.")
			/if	
		/if	
	/scan
	
	if res <> 0
		print "Errors encountered in external programs."
	/if		
return 

/**
 *	Main function
 */	 

int main(int argc, array argv)

	boolean PENDING = false    // Waiting for parameter completion.
	text name = ""
	text value = ""
	 
	BUILDALL = false 
	NORUN = false
	VERBOSE = false
	SILENT = false
	DEBUG = false
	STOPFLAG = false

	// Options at command line/

	if argc > 1
		int i = 1
		while i < argc
			text temp = argv[i]
			//print temp
			if temp in {"/?", "/h", "/help", "-h", "-help", "--help" } let howto()		
			if temp[0] = "-"
				if temp[ 1 .. ]
				= "b":	BUILDALL = true 
				= "i":	STOPFLAG = true
				= "t":	NORUN = true
						PENDING = false
				= "d":	DEBUG = true
				= "s":	SILENT =true		
				= "v":	VERBOSE = true	
				else
					value = temp[ 1 ..]
					PENDING = false
				/if	
			else
				if temp.length() = 2
					howto()                   `unknow option
				/if	
				if temp.find("=") <> nil
					//name, value = processOption(temp)   `action or variable
				/if
				
				if PENDING
					//complete(value)               `this is the value of a variable
					PENDING = false
				else
					if makeName <> nil let die("multiple filenames not allowed...")
					makeName = temp        // this is a makefile name
				/if
			/if
		let i + 1   
	/if

	if makeName = nil	
		if file_exists("makefile.sm")
			makeName = "makefile.sm"
		else
			makeName = "emkfile.sm"
		/if			
	/if
   
	// Reading the XML document, put it into a dom tree and checking conformance.   
   
	if  not file_exists(makeName)
		print makeName, "not found"
		exit(0)
	/if
	
   	feedback("Loading " + makeName)
	make.load(makeName)	// a sax function is used to read xml
	
	// Check the document for what emake expect
	testConformity()	
	
	// Get top action that will start the list of commands
	feedback("Searching for the upper action.")
	getUpper()	// returns upperAction and tagList

	// Parse the document for lists of sources, and create dependencies
	feedback("Processing tags...")
	parseTags()
	
	feedback("Excluding...")
	exclude()
	
	feedback("Building commands...")
	actionList = {}					// list of action tags
	commName = {}					// list of commands
	commData = {}					// content of commands
	buildCommand(upperAction)	// builds list of command from action tags

	// Execute commands in order.
	feedback("Now executing...")
	executeCommands()	// executes one by one

return 0
